# -*- coding: utf-8 -*-

# Octopasty is an Asterisk AMI proxy

# Copyright (C) 2011  Jean Schurger <jean@schurger.org>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from select import select
from utils import bigtime
from time import time
from copy import copy

from internal import handle_action


def readall(self):
    reading = self.ami_files + self.client_files
    if len(reading) > 0:
        # it's a timeout, that means that 0 != None
        timeout = self.out_queue.empty() and 1 or 0
        to_read, _, _ = select(reading, [], [], timeout)
        for f in to_read:
            peer = self.find_peer_from_file(f)
            peer.handle_line()


def sendall(self):
    if not self.out_queue.empty():
        outgoing = list(self.out_queue.queue)
        self.out_queue.queue.clear()
        for packet in outgoing:
            if packet.dest == '__internal__':
                handle_action(self, packet)
            else:
                dest = self.get_peer(packet.dest)
                sent = dest.send(packet)
                if sent:
                    if sent in self.flow:
                        # then it was an answer
                        self.flow.pop("%s" % sent)
                    else:
                        # then it was a query
                        self.flow["%s" % sent] = packet.emiter
                else:
                    # They it was an event
                    pass


def burials(self):
    apocalypse = bigtime(-10)
    # Remove old locks
    for peer in self.peers:
        if peer and peer.logged and peer.locked < apocalypse:
            peer.locked = 0


# The nervous central
# Thanks to Derek for the awesome function name !
def squirm(self):
    apocalypse = time() - 10
    if not self.in_queue.empty():
        incoming = list(self.in_queue.queue)
        self.in_queue.queue.clear()
        for packet in incoming:
            if packet.timestamp < apocalypse:
                # not puting in the queue and just continuing
                # makes the packet go nowhere
                if packet.locked:
                    dest = dest = self.flow.get("%s" % packet.locked)
                    if dest != '__internal__':
                        peer = self.get_peer(dest)
                        # HERE
            if packet.emiter == '__internal__':
                peer = self.get_peer(packet.dest)
                if peer and peer.available:
                    self.out_queue.put(packet)
                else:
                    self.in_queue.put(packet)
            else:
                if packet.locked:
                    if packet.emiter in self.connected_servers:
                        emiter = self.get_peer(packet.emiter)
                        emiter.locked = 0
                        dest = self.flow.get("%s" % packet.locked)
                        if dest:
                            if dest == '__internal__':
                                handle_action(self, packet)
                            else:
                                peer = self.get_peer(dest)
                                if peer and peer.locked == packet.locked \
                                       and peer.logged:
                                    packet.dest = dest
                                    self.out_queue.put(packet)
                                else:
                                    self.in_queue.put(packet)
                    if packet.emiter in self.connected_clients:
                        client = self.get_peer(packet.emiter)
                        if client.logged:
                            cid = client.id.split('_')[0]
                            cu = self.config['users'].get(cid)
                            if cu:
                                cs = cu.get('server')
                                if cs:
                                    peer = self.get_peer(cs)
                                    if peer and peer.available and peer.logged:
                                        packet.dest = cs
                                        self.out_queue.put(packet)
                                    else:
                                        self.in_queue.put(packet)
                        else:
                            packet.dest = '__internal__'
                            self.out_queue.put(packet)
                else:
                    # packet not locked
                    if packet.emiter in self.connected_servers:
                        # broadcast events
                        ami = self.get_peer(packet.emiter)
                        if ami.logged:
                            for client in self.connected_clients:
                                peer = self.get_peer(client)
                                if peer.binded_server == packet.emiter and \
                                   peer.wants_events:
                                    p = copy(packet)
                                    p.dest = client
                                    self.out_queue.put(p)
                        else:
                            # Do not remove that comment and that case
                            # it might be logged one day
                            # ami not logged, and no command before
                            # i wonder what can it be
                            pass



# -*- coding: utf-8 -*-

# Octopasty is an Asterisk AMI proxy

# Copyright (C) 2011  Jean Schurger <jean@schurger.org>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import socket
from threading import Thread
from time import sleep, time
from asterisk import Action

from utils import Packet, bigtime


class ServerThread(Thread):

    def __init__(self, octopasty, channel, details):
        Thread.__init__(self)
        self.octopasty = octopasty
        self.channel = channel
        self.details = details
        self.action = None
        self.logged = False
        self.id = 'unknown_%d' % bigtime()
        self.locked = 0
        self.wants_events = False
        self.binded_server = None

    def disconnect(self):
        if self.channel:
            self.channel.close()
            self.channel = None
        if self.id in self.octopasty.clients:
            self.octopasty.clients.pop(self.id)
        self.connected = False
        self.file = None

    def run(self):
        self.file = self.channel.makefile()
        self.connected = True
        self.file.write("Asterisk Call Manager/1.1\n")
        self.file.flush()
        self.octopasty.clients.update({self.id: self})

    def send(self, packet):
        released_lock = None
        if self.locked:
            if packet.locked == self.locked:
                # it's for me
                self.file.write(packet.packet)
                self.file.flush()
                released_lock = self.locked
                self.locked = 0
            else:
                # just discard, may be too old
                pass
        else:
            # on let events go
            if not packet.locked:
                self.file.write(packet.packet)
                self.file.flush()
            else:
                # humm, why we get that ??
                pass
        return released_lock

    def handle_line(self):
        can_read = False
        if self.channel:
            can_read = True
            self.channel.setblocking(0)
        while can_read and self.file:
            try:
                line = self.file.readline()
                if len(line) == 0:
                    self.disconnect()
                else:
                    line = line.strip()
                    # if locked, we are waiting for a result
                    if not self.locked:
                        if line.startswith('Action:'):
                            self.action = Action(line.replace('Action:',
                                                              '').strip())
                        elif line == '':
                            if self.action:
                                self.locked = bigtime()
                                self.push(self.action)
                                self.action = None
                        else:
                            if ':' in line:
                                k = line.split(':')[0]
                                v = ':'.join(line.split(':')[1:]).lstrip()
                            if self.action:
                                self.action.add_parameters({k: v})
            except socket.error:
                can_read = False
        if self.channel:
            self.channel.setblocking(1)

    def push(self, packet):
        p = dict(emiter=self.id, locked=self.locked,
                 timestamp=time(), packet=packet)
        self.octopasty.in_queue.put(Packet(p))

    def _get_available(self):
        return self.connected and not self.locked
    available = property(_get_available)

    def _get_uid(self):
        return self.id
    uid = property(_get_uid)


class MainListener(Thread):

    def __init__(self, octopasty):
        Thread.__init__(self)
        self.octopasty = octopasty
        self.start()

    def run(self):
        while True:
            try:
                server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

                server.bind((self.octopasty.config.get('bind_address'),
                             int(self.octopasty.config.get('bind_port'))))
                print "Listening on %s %s" % \
                      (self.octopasty.config.get('bind_address'),
                       self.octopasty.config.get('bind_port'))
                server.listen(5)
                while True:
                    channel, details = server.accept()
                    st = ServerThread(self.octopasty, channel, details)
                    st.start()
            except socket.error, e:
                print "%s Try to bind on %s %s" % \
                      (e, self.octopasty.config.get('bind_address'),
                       self.octopasty.config.get('bind_port'))
                sleep(10)



# -*- coding: utf-8 -*-

# Octopasty is an Asterisk AMI proxy

# Copyright (C) 2011  Jean Schurger <jean@schurger.org>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import socket
from threading import Thread
from time import time
from utils import Packet, bigtime

from asterisk import Login
from asterisk import Event, Response


class AMIClient(Thread):
    def __init__(self, octopasty, server, parameters):
        Thread.__init__(self)
        self.octopasty = octopasty
        self.server = server
        self.socket = None
        self.host = parameters.get('host')
        self.port = parameters.get('port')
        self.user = parameters.get('user')
        self.password = parameters.get('password')
        self.connected = False
        self.file = None
        self.response = None
        self.event = None
        self.logged = False
        self.locked = False

    def disconnect(self):
        if self.socket:
            self.socket.close()
            self.socket = None
        if self.server in self.octopasty.amis:
            self.octopasty.amis.pop(self.server)
        self.connected = False
        self.file = None

    def send(self, packet):
        if not self.locked:
            self.file.write(packet.packet)
            self.file.flush()
            #self.locked = int(time() * 10000000)
            self.locked = packet.locked
            return self.locked
        else:
            return None

    def login(self):
        self.push(Login(self.user, self.password),
                  emiter='__internal__', dest=self.server,
                  locked=bigtime())

    def run(self):
        try:
            s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            s.connect((self.host, int(self.port)))
            self.socket = s
            self.octopasty.amis.update({self.server: self})
            self.connected = True
            self.file = self.socket.makefile(bufsize=42)
        except socket.error:
            pass

    def handle_line(self):
        can_read = False
        if self.socket:
            can_read = True
            self.socket.setblocking(0)
        while can_read and self.file:
            try:
                line = self.file.readline()
                if len(line) == 0:
                    self.disconnect()
                else:
                    line = line.strip()
                    if line.startswith('Asterisk Call Manager'):
                        self.login()
                        return
                    if self.locked and line.startswith('Response:'):
                        self.response = Response(line.replace('Response:',
                                                              '').strip())
                        self.event = None
                    elif line.startswith('Event:'):
                        self.event = Event(line.replace('Event:', '').strip())
                        self.response = None
                    elif line == '':
                        if self.response:
                            self.push(self.response)
                            self.response = None
                        elif self.event:
                            self.push(self.event)
                            self.event = None
                    else:
                        if ':' in line:
                            k = line.split(':')[0]
                            v = ':'.join(line.split(':')[1:]).lstrip()
                        else:
                            k = line
                            v = None
                        if self.response:
                            self.response.add_parameters({k: v})
                        if self.event:
                            self.event.add_parameters({k: v})
            except socket.error:
                can_read = False
        if self.socket:
            self.socket.setblocking(1)

    def push(self, packet, emiter=None, dest=None, locked=0):
        p = dict(emiter=emiter or self.server,
                 locked=locked or self.locked, timestamp=time(),
                 packet=packet)
        if dest:
            p['dest'] = dest
        self.octopasty.in_queue.put(Packet(p))

    def _get_available(self):
        return self.connected and not self.locked
    available = property(_get_available)

    def _get_uid(self):
        return self.server
    uid = property(_get_uid)



# -*- coding: utf-8 -*-

# Octopasty is an Asterisk AMI proxy

# Copyright (C) 2011  Jean Schurger <jean@schurger.org>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


class Packet(object):
    _name_ = 'Packet'

    def __init__(self, name=None, parameters=None):
        self.parameters = dict()
        self.name = name
        if self.name and parameters:
            self.add_parameters(parameters)

    def add_parameters(self, parameters):
        self.parameters.update(parameters)

    def __repr__(self):
        out = "Action: %s\n" % self.name
        for p in self.parameters:
            if self.parameters[p] == None:
                out += "%s\n" % p
                pass
            else:
                out += "%s: %s\n" % (p, self.parameters.get(p))
        out += "\n"
        return out


class Command(Packet):
    _name_ = 'Command'
    pass


class Login(Command):
    _name_ = 'Login'

    def __init__(self, login, password):
        Command.__init__(self, 'login', dict(Username=login,
                                             Secret=password,
                                             Events='on'))


class Answer(Packet):
    _name_ = 'Answer'


class Event(Answer):
    _name_ = 'Event'


class Response(Packet):
    _name_ = 'Response'


class Action(Answer):
    _name_ = 'Action'


class Success(Action):
    _name_ = 'Success'

    def __init__(self, parameters):
        Action.__init__(self, 'Success', parameters)


class Error(Action):
    _name_ = 'Error'

    def __init__(self, parameters):
        Action.__init__(self, 'Error', parameters)
